import pytest
import spacy
from thinc.api import Model
from ..layers import TransformerModel
from ..data_classes import FullTransformerBatch
from ..span_getters import get_doc_spans


MODEL_NAMES = [
    "distilbert-base-uncased",
    "hf-internal-testing/tiny-random-gpt2",
    "hf-internal-testing/tiny-random-xlnet",
]


@pytest.fixture
def nlp():
    return spacy.blank("en")


@pytest.fixture
def docs(nlp):
    texts = ["the cat sat on the mat.", "hello world."]
    return [nlp(text) for text in texts]


@pytest.fixture(scope="module", params=MODEL_NAMES)
def name(request):
    return request.param


@pytest.fixture(scope="module", params=[True, False])
def output_attentions(request):
    return request.param


@pytest.fixture(scope="module", params=[True, False])
def output_hidden_states(request):
    return request.param


@pytest.fixture(scope="module")
def trf_model(name, output_attentions, output_hidden_states):
    if "gpt2" in name:
        model = TransformerModel(
            name,
            get_doc_spans,
            {"use_fast": True, "pad_token": "<|endoftext|>"},
            {
                "output_attentions": output_attentions,
                "output_hidden_states": output_hidden_states,
            },
        )

    else:
        # test slow tokenizers with distilbert-base-uncased (parameterizing
        # for all models blows up the memory usage during the test suite)
        if name == "distilbert-base-uncased":
            use_fast = False
        else:
            use_fast = True
        model = TransformerModel(
            name,
            get_doc_spans,
            {"use_fast": use_fast},
            {
                "output_attentions": output_attentions,
                "output_hidden_states": output_hidden_states,
            },
        )
    model.initialize()
    return model


def test_model_init(name, trf_model):
    assert isinstance(trf_model, Model)
    if name == "distilbert-base-uncased":
        assert not trf_model.tokenizer.is_fast
    else:
        assert trf_model.tokenizer.is_fast


def test_model_predict(nlp, docs, trf_model):
    outputs = trf_model.predict(docs)
    shape = outputs.model_output.last_hidden_state.shape
    if trf_model.transformer.config.output_attentions is True:
        assert outputs.model_output.attentions is not None
        assert all([t.shape[0] == shape[0] for t in outputs.model_output.attentions])
    else:
        assert outputs.model_output.attentions is None
    if trf_model.transformer.config.output_hidden_states is True:
        assert outputs.model_output.hidden_states is not None
        assert all([t.shape[0] == shape[0] for t in outputs.model_output.hidden_states])
    else:
        assert outputs.model_output.hidden_states is None
    assert isinstance(outputs, FullTransformerBatch)

    # for a fast tokenizer check that all non-special wordpieces are aligned
    # (which is not necessarily true for the slow tokenizers)
    if trf_model.tokenizer.is_fast:
        outputs = trf_model.predict([nlp.make_doc("\tÁaaa  \n\n")])
        aligned_wps = outputs.align.data.flatten()
        for i in range(len(outputs.wordpieces.strings[0])):
            if (
                outputs.wordpieces.strings[0][i]
                not in trf_model.tokenizer.all_special_tokens
            ):
                assert i in aligned_wps



from typing import Callable
from functools import partial
import copy

import torch
from transformers import AutoModelForSequenceClassification
from transformers.models.distilbert.modeling_distilbert import (
    DistilBertForSequenceClassification,
)
from transformers.modeling_outputs import SequenceClassifierOutput

import spacy
from thinc.api import Model

from spacy_transformers.data_classes import HFObjects, WordpieceBatch
from spacy_transformers.layers.hf_wrapper import HFWrapper
from spacy_transformers.layers.transformer_model import _convert_transformer_inputs
from spacy_transformers.layers.transformer_model import _convert_transformer_outputs
from spacy_transformers.layers.transformer_model import forward
from spacy_transformers.layers.transformer_model import huggingface_from_pretrained
from spacy_transformers.layers.transformer_model import huggingface_tokenize
from spacy_transformers.layers.transformer_model import set_pytorch_transformer
from spacy_transformers.span_getters import get_strided_spans


def test_model_for_sequence_classification():
    # adapted from https://github.com/KennethEnevoldsen/spacy-wrap/
    class ClassificationTransformerModel(Model):
        def __init__(
            self,
            name: str,
            get_spans: Callable,
            tokenizer_config: dict = {},
            transformer_config: dict = {},
            mixed_precision: bool = False,
            grad_scaler_config: dict = {},
        ):
            hf_model = HFObjects(None, None, None, tokenizer_config, transformer_config)
            wrapper = HFWrapper(
                hf_model,
                convert_inputs=_convert_transformer_inputs,
                convert_outputs=_convert_transformer_outputs,
                mixed_precision=mixed_precision,
                grad_scaler_config=grad_scaler_config,
                model_cls=AutoModelForSequenceClassification,
            )
            super().__init__(
                "clf_transformer",
                forward,
                init=init,
                layers=[wrapper],
                dims={"nO": None},
                attrs={
                    "get_spans": get_spans,
                    "name": name,
                    "set_transformer": set_pytorch_transformer,
                    "has_transformer": False,
                    "flush_cache_chance": 0.0,
                },
            )

        @property
        def tokenizer(self):
            return self.layers[0].shims[0]._hfmodel.tokenizer

        @property
        def transformer(self):
            return self.layers[0].shims[0]._hfmodel.transformer

        @property
        def _init_tokenizer_config(self):
            return self.layers[0].shims[0]._hfmodel._init_tokenizer_config

        @property
        def _init_transformer_config(self):
            return self.layers[0].shims[0]._hfmodel._init_transformer_config

        def copy(self):
            """
            Create a copy of the model, its attributes, and its parameters. Any child
            layers will also be deep-copied. The copy will receive a distinct `model.id`
            value.
            """
            copied = ClassificationTransformerModel(self.name, self.attrs["get_spans"])
            params = {}
            for name in self.param_names:
                params[name] = self.get_param(name) if self.has_param(name) else None
            copied.params = copy.deepcopy(params)
            copied.dims = copy.deepcopy(self._dims)
            copied.layers[0] = copy.deepcopy(self.layers[0])
            for name in self.grad_names:
                copied.set_grad(name, self.get_grad(name).copy())
            return copied

    def init(model: ClassificationTransformerModel, X=None, Y=None):
        if model.attrs["has_transformer"]:
            return
        name = model.attrs["name"]
        tok_cfg = model._init_tokenizer_config
        trf_cfg = model._init_transformer_config
        hf_model = huggingface_from_pretrained(
            name, tok_cfg, trf_cfg, model_cls=AutoModelForSequenceClassification
        )
        model.attrs["set_transformer"](model, hf_model)
        tokenizer = model.tokenizer
        texts = ["hello world", "foo bar"]
        token_data = huggingface_tokenize(tokenizer, texts)
        wordpieces = WordpieceBatch.from_batch_encoding(token_data)
        model.layers[0].initialize(X=wordpieces)

    model = ClassificationTransformerModel(
        "sgugger/tiny-distilbert-classification",
        get_spans=partial(get_strided_spans, window=128, stride=96),
    )
    model.initialize()

    assert isinstance(model.transformer, DistilBertForSequenceClassification)
    nlp = spacy.blank("en")
    doc = nlp.make_doc("some text")
    assert isinstance(model.predict([doc]).model_output, SequenceClassifierOutput)

    b = model.to_bytes()
    model_re = ClassificationTransformerModel(
        "sgugger/tiny-distilbert-classification",
        get_spans=partial(get_strided_spans, window=128, stride=96),
    ).from_bytes(b)
    assert isinstance(model_re.transformer, DistilBertForSequenceClassification)
    assert isinstance(model_re.predict([doc]).model_output, SequenceClassifierOutput)
    assert torch.equal(
        model.predict([doc]).model_output.logits,
        model_re.predict([doc]).model_output.logits,
    )
    # Note that model.to_bytes() != model_re.to_bytes(), but this is also not
    # true for the default models.



from functools import partial

import pytest
import spacy
from spacy.training import Example
from spacy.training.initialize import init_nlp
from spacy.util import CONFIG_SECTION_ORDER
from spacy.language import DEFAULT_CONFIG
from thinc.config import Config


TRAIN_TAGGER_DATA = [
    ("I like green eggs", {"tags": ["N", "V", "J", "N"]}),
    ("Eat blue ham", {"tags": ["V", "J", "N"]}),
]


cfg_string = """
    [nlp]
    lang = "en"
    pipeline = ["custom_transformer","tagger"]

    [components]

    [components.tagger]
    factory = "tagger"

    [components.tagger.model]
    @architectures = "spacy.Tagger.v1"
    nO = null

    [components.tagger.model.tok2vec]
    @architectures = "spacy-transformers.TransformerListener.v1"
    grad_factor = 1.0
    upstream = "custom_transformer"

    [components.tagger.model.tok2vec.pooling]
    @layers = "reduce_mean.v1"

    [components.custom_transformer]
    factory = "transformer"

    [corpora]
    @readers = toy_tagger_data.v1
    
    [initialize]
    
    [initialize.components]
    
    [initialize.components.tagger]
    labels = ["LABEL"]
    """


@pytest.mark.parametrize("config_string", [cfg_string])
def test_init_nlp(config_string):
    @spacy.registry.readers.register("toy_tagger_data.v1")
    def read_tagger_data():
        def parse_data(nlp, index):
            ex = TRAIN_TAGGER_DATA[index]
            yield Example.from_dict(nlp.make_doc(ex[0]), ex[1])

        return {
            "train": partial(parse_data, index=0),
            "dev": partial(parse_data, index=1),
        }

    config = spacy.util.load_config_from_str(config_string, interpolate=False)
    config = Config(DEFAULT_CONFIG, section_order=CONFIG_SECTION_ORDER).merge(config)
    nlp = init_nlp(config, use_gpu=False)
    assert nlp is not None

    tagger = nlp.get_pipe("tagger")
    transformer = nlp.get_pipe("custom_transformer")
    tagger_trf = tagger.model.get_ref("tok2vec").layers[0]
    assert tagger_trf.upstream_name == "custom_transformer"
    assert transformer.listeners[0] == tagger_trf



from typing import Callable, Iterable, List
from functools import partial
from spacy.tokens import Doc, Span

from .util import registry

SpannerT = Callable[[List[Doc]], List[List[Span]]]


def get_strided_spans(
    docs: Iterable[Doc], window: int, stride: int
) -> List[List[Span]]:
    spans: List[List[Span]] = []
    for doc in docs:
        start = 0
        spans.append([])
        for i in range(len(doc) // stride):
            spans[-1].append(doc[start : start + window])
            if (start + window) >= len(doc):
                break
            start += stride
        else:
            if start < len(doc):
                spans[-1].append(doc[start:])
    return spans


@registry.span_getters("spacy-transformers.strided_spans.v1")  # type: ignore
def configure_strided_spans(window: int, stride: int) -> SpannerT:
    """
    Set the 'window' and 'stride' options for getting strided spans.

    If you set the window and stride to the same value, the spans will cover
    each token once. Setting 'stride' lower than 'window' will allow for an
    overlap, so that some tokens are counted twice. This can be desirable,
    because it allows all tokens to have both a left and right context.
    """
    return partial(get_strided_spans, window=window, stride=stride)


def get_sent_spans(docs: Iterable[Doc]) -> List[List[Span]]:
    return [list(doc.sents) for doc in docs]


@registry.span_getters("spacy-transformers.sent_spans.v1")  # type: ignore
def configure_get_sent_spans() -> Callable:
    """
    Create a `span_getter` that uses sentence boundary markers to extract
    the spans. This requires sentence boundaries to be set, and may result
    in somewhat uneven batches, depending on the sentence lengths. However,
    it does provide the transformer with more meaningful windows to attend over.
    """
    return get_sent_spans


def get_doc_spans(docs: Iterable[Doc]) -> List[List[Span]]:
    return [[doc[:]] for doc in docs]


@registry.span_getters("spacy-transformers.doc_spans.v1")  # type: ignore
def configure_get_doc_spans() -> Callable:
    """
    Create a `span_getter` that uses the whole document as its spans. This is
    the best approach if your `Doc` objects already refer to relatively short
    texts.
    """
    return get_doc_spans


__all__ = [
    "get_sent_spans",
    "get_doc_spans",
    "configure_get_doc_spans",
    "configure_get_sent_spans",
    "configure_strided_spans",
]



from typing import Callable, List
from spacy.tokens import Doc

from .util import registry
from .data_classes import FullTransformerBatch


def null_annotation_setter(docs: List[Doc], trf_data: FullTransformerBatch) -> None:
    """Set no additional annotations on the Doc objects."""
    pass


@registry.annotation_setters("spacy-transformers.null_annotation_setter.v1")  # type: ignore
def configure_null_annotation_setter() -> Callable[
    [List[Doc], FullTransformerBatch], None
]:
    return null_annotation_setter


__all__ = ["null_annotation_setter", "configure_null_annotation_setter"]



from typing import Dict, List, Tuple, Callable, Optional
from spacy.tokens import Span, Token
from thinc.api import Ops
from thinc.types import Ragged, Floats2d, Ints2d

def apply_alignment(
    ops: Ops, align: Ragged, X: Floats2d
) -> Tuple[Ragged, Callable]: ...
def get_token_positions(spans: List[Span]) -> Dict[Token, int]: ...
def get_alignment_via_offset_mapping(
    spans: List[Span],
    offset_mapping: Ints2d,
) -> Ragged: ...
def get_alignment(
    spans: List[Span],
    wordpieces: List[List[str]],
    special_tokens: Optional[List[str]] = None,
) -> Ragged: ...
def get_span2wp_from_offset_mapping(
    span: Span,
    wp_char_offsets: Tuple[int],
) -> List[List[int]]: ...



import pytest
from typing import List
import numpy
from spacy.tokens import Doc
from spacy.vocab import Vocab
from thinc.api import NumpyOps
from thinc.types import Ragged
from ..align import get_alignment, apply_alignment
from ..align import get_span2wp_from_offset_mapping


def get_ragged(ops, nested: List[List[int]]):
    nested = [ops.asarray(x) for x in nested]
    return Ragged(ops.flatten(nested), ops.asarray([len(x) for x in nested]))


def get_spans(word_seqs):
    vocab = Vocab()
    docs = [Doc(vocab, words=words) for words in word_seqs]
    return [doc[:] for doc in docs]


def flatten_strings(words1, words2):
    flat1 = []
    flat2 = []
    for seq in words1:
        flat1.extend(seq)
    stride = max((len(seq) for seq in words2), default=0)
    for seq in words2:
        flat2.extend(seq)
        flat2.extend([""] * (stride - len(seq)))
    return flat1, flat2


@pytest.mark.parametrize(
    "words1,words2",
    [
        ([["a", "b"]], [["a", "b"]]),
        ([["ab"]], [["a", "b"]]),
        ([["a", "b"]], [["ab"]]),
        ([["ab", "c"]], [["a", "bc"]]),
        ([["ab", "cd"]], [["a", "bc", "d"]]),
    ],
)
def test_alignments_match(words1, words2):
    spans = get_spans(words1)
    align = get_alignment(spans, words2)
    unique_tokens = set()
    for span in spans:
        for token in span:
            unique_tokens.add((id(token.doc), token.idx))
    assert len(unique_tokens) == align.lengths.shape[0]
    flat_words1, flat_words2 = flatten_strings(words1, words2)
    for i, word in enumerate(flat_words1):
        wp_word = "".join([flat_words2[int(j)] for j in align[i].data])
        if len(word) < len(wp_word):
            assert word in wp_word
        elif len(word) > len(wp_word):
            assert wp_word in word
        else:
            assert word == wp_word


@pytest.mark.parametrize(
    "nested_align,X_cols",
    [
        ([[0, 1, 2], [3], [4]], 4),
        ([[], [1], [1], [2]], 2),
        ([[0, 1], [1, 2], [], [4]], 2),
    ],
)
def test_apply_alignment(nested_align, X_cols):
    ops = NumpyOps()
    align = get_ragged(ops, nested_align)
    X_shape = (align.data.max() + 1, X_cols)
    X = ops.alloc2f(*X_shape)
    Y, get_dX = apply_alignment(ops, align, X)
    assert isinstance(Y, Ragged)
    assert Y.data.shape[0] == align.data.shape[0]
    assert Y.lengths.shape[0] == len(nested_align)
    dX = get_dX(Y)
    assert dX.shape == X.shape


@pytest.mark.parametrize(
    # fmt: off
    # roberta-base offset_mapping and expected alignment
    "words,offset_mapping,alignment",
    [
        (
            ["Áaaa"],
            numpy.asarray([(0, 0), (0, 1), (0, 1), (1, 4), (0, 0)], dtype="i"),
            [[1, 2, 3]],
        ),
        (
            ["INGG", "á", "aäa"],
            numpy.asarray([(0, 0), (0, 3), (3, 4), (5, 6), (5, 6), (7, 8), (8, 9), (9, 10), (0, 0)], dtype="i"),
            [[1, 2], [3, 4], [5, 6, 7]],
        ),
    ],
    # fmt: on
)
def test_offset_alignment(words, offset_mapping, alignment):
    spans = get_spans([words])
    result = get_span2wp_from_offset_mapping(spans[0], offset_mapping)
    assert all(sorted(r) == a for r, a in zip(result, alignment))



from thinc.api import chain
from .split_trf import split_trf_batch


def replace_listener(model):
    return chain(model, split_trf_batch())


def replace_listener_cfg(tok2vec_model_cfg, listener_model_cfg):
    result = tok2vec_model_cfg.copy()
    if (
        "TransformerModel" in tok2vec_model_cfg["@architectures"]
        and "TransformerListener" in listener_model_cfg["@architectures"]
    ):
        result["@architectures"] = "spacy-transformers.Tok2VecTransformer.v3"
        for key in ["pooling", "grad_factor"]:
            if key in listener_model_cfg and key not in result:
                result[key] = listener_model_cfg[key]
    return result
